/*
** Multi-object block file storage with recursively paged headers.
**
** [Header:16]
** {AllocTable Head: <Depth:1> <Block:4> } (initial 0, 0)
** {ObjectTable Head: <Depth:1> <Block:4> } (initial 0, 1)
**
** [Object Table]
** { <Pointer:4> <Length:4> }*
**
** [Data Block]
** <Data:Z-4> <Next:4>
*/

use pack::prelude::*;

use std::{
    fs::File,
    io::{Read, Seek, SeekFrom, Write},
    path::Path,
};

const HEADER_SIZE :usize = 16;

const F_NOVACANCY :u32 = 1 << 31;

enum Operation {
    None,
    SetVacant,
    SetOccupied,
}

struct OperResult {
    pub result:usize,
    pub operation:Operation,
}

pub struct BlockFile<const Z:usize> {
    file:File,
}
impl<const Z:usize> BlockFile<Z> {
    pub fn open<P:AsRef<Path>>(path:P) -> Result<Self, std::io::Error>
    {
        if path.as_ref().exists() {

            // Open existing file.
            match File::options()
            .read(true)
            .write(true)
            .open(path) {
                Ok(file) => {
                    Ok(Self { file })
                }
                Err(error) => Err(error),
            }
        } else {

            // Create and initialize new file.
            match File::options()
            .create_new(true)
            .read(true)
            .write(true)
            .open(path) {
                Ok(file) => {
                    Self { file }.init()
                }
                Err(error) => Err(error),
            }
        }
    }

    fn init(mut self) -> Result<Self, std::io::Error>
    // Write initial headers to the file.
    //
    {
        /*
        ** Header size and first two blocks are initialized for
        ** first pages of allocation table and object table.
        */
        let mut data = vec![0u8; HEADER_SIZE + (Z * 4)];
        data[0] = 1;
        data[5] = 1;
        data[6] = 2;
        data[HEADER_SIZE] = 1;
        data[HEADER_SIZE + Z] = 0x0F;
        data[HEADER_SIZE + (Z * 2)] = 3;

        for i in 0..Self::table_size() {
            let index = HEADER_SIZE + (Z * 3) + (i * 8);
            let pack_pointer = (1 + i as u32).pack();

            data[index] = pack_pointer[0];
            data[index + 1] = pack_pointer[1];
            data[index + 2] = pack_pointer[2];
            data[index + 3] = pack_pointer[3];
        }

        self.file.write(&data)?;
        Ok(self)
    }

    pub fn insert(&mut self, data:&[u8]) -> Result<usize, std::io::Error>
    // Acquire a new object id and write its data to new allocations.
    //
    {

        // Allocate storage blocks
        let block_count = Self::block_count_from_size(data.len());
        //println!("block_count {}", block_count);

        let blocks = self.allocate(block_count.max(1))?;
        //println!("blocks {}", blocks.len());

        // Get object id
        let id = self.acquire_object(blocks[0], data.len())?;

        //println!("obj_id {}", id);

        // Write data to storage blocks
        let mut data_index = 0;
        let mut block_data = vec![0u8; Z];

        for block_index in 0..block_count {
            block_data.fill(0);

            // Copy data slice to buffer
            for b in 0..Self::data_size().min(data.len() - (block_index * Self::data_size())) {
                block_data[b] = data[data_index + b];
            }
            data_index += Self::data_size();

            // Write pointer to next block to end of buffer
            if block_index < block_count - 1 {
                let pack_next = (blocks[block_index + 1] as u32).pack();
                block_data[Z - 4] = pack_next[0];
                block_data[Z - 3] = pack_next[1];
                block_data[Z - 2] = pack_next[2];
                block_data[Z - 1] = pack_next[3];
            }

            self.write_block(blocks[block_index], &block_data)?;
        }

        Ok(id)
    }

    pub fn update(&mut self, id:usize, data:&[u8]) -> Result<(), std::io::Error>
    // Write new data to an object expanding its existing allocation.
    //
    {
        // Get first block and data size
        let (mut block_id, _) = self.get_object(id)?;

        let required_blocks = Self::block_count_from_size(data.len());
        let mut blocks = Vec::new();

        // Count blocks
        while block_id != 0 {
            let next = self.read_block_pointer(block_id)?;
            blocks.push(block_id);
            block_id = next;
        }

        if blocks.len() > required_blocks {
            // Free excess blocks
            self.release(&blocks[required_blocks..])?;
            blocks.resize(required_blocks, 0);

        } else if blocks.len() < required_blocks {
            // Allocate additional blocks
            let allocated_blocks = self.allocate(required_blocks - blocks.len())?;
            blocks.extend_from_slice(&allocated_blocks);
        }

        // Write data to storage blocks
        let mut data_index = 0;
        let mut block_data = vec![0u8; Z];
        
        for block_index in 0..blocks.len() {
            block_data.fill(0);

            // Copy data slice to buffer
            for b in 0..Self::data_size().min(data.len() - (block_index * Self::data_size())) {
                block_data[b] = data[data_index + b];
            }
            data_index += Self::data_size();

            // Write pointer to next block to end of buffer
            if block_index < blocks.len() - 1 {
                let pack_next = (blocks[block_index + 1] as u32).pack();
                block_data[Z - 4] = pack_next[0];
                block_data[Z - 3] = pack_next[1];
                block_data[Z - 2] = pack_next[2];
                block_data[Z - 1] = pack_next[3];
            }

            self.write_block(blocks[block_index], &block_data)?;
        }

        Ok(())
    }

    pub fn remove(&mut self, _id:usize) -> Result<(), std::io::Error>
    // Remove the object with the specified identifier and release its allocation.
    //
    {
        Ok(())
    }

    pub fn get(&self, id:usize) -> Result<Vec<u8>, std::io::Error>
    // Return the object with the specified identifier.
    //
    {
        let mut block_data = [0u8; Z];

        // Get first block and data size
        let (mut block_id, size) = self.get_object(id)?;
        let mut data = Vec::new();

        // Read blocks until size is full
        while block_id != 0 {
            self.read_block(block_id, &mut block_data)?;
            let next_block = u32::unpack(&block_data, &mut (Z - 4)).unwrap_or_default();
            
            //println!("size {} len {}", size, data.len());

            let data_length = if next_block != 0 {
                Z - 4
            } else {
                size - data.len()
            };

            data.extend_from_slice(&block_data[0..data_length]);

            block_id = next_block;
        }
        
        Ok(data)
    }

    pub fn ids(&self) -> Result<Vec<usize>, std::io::Error>
    // Traverses object list and returns allocated ids.
    //
    {
        let output = Vec::new();

        //

        Ok(output)
    }

    fn allocate(&mut self, count:usize) -> Result<Vec<u32>, std::io::Error>
    // Mark as allocated and return the next available N blocks.
    //
    {
        let mut b8 = [0u8; 1];
        let mut b32 = [0u8; 4];

        // Read allocation table root block and depth from file.
        self.file.seek(SeekFrom::Start(0))?;
        self.file.read_exact(&mut b8)?;
        self.file.read_exact(&mut b32)?;

        let root_block = u32::unpack(&b32, &mut 0).unwrap_or_default();

        let mut blocks = Vec::new();

        // Acquire next available block in table.
        self.allocate_traverse(
            root_block,
            b8[0] as u32,
            0,
            true,
            count,
            &mut blocks,
        )?;

        Ok(blocks)
    }

    fn allocate_traverse(
        &mut self,
        block_id:u32,
        depth:u32,
        basis:u32,
        is_root:bool,
        count:usize,
        blocks:&mut Vec<u32>,
    ) -> Result<Operation, std::io::Error>
    // Search allocation page tables for next available child page.
    // Mark blocks allocated until requested number is acquired.
    //
    {
        //println!("allocate_traverse()");

        let mut block_data = [0u8; Z];
        self.read_block(block_id, &mut block_data)?;
        let mut write_block = false;

        let mut operation = Operation::None;

        //println!("alc node {}, depth {}", block_id, depth);

        // Search table for first vacant or unallocated child.
        if depth > 0 {

            /* Check each cell for 
            ** Vacant cells have a MSB of 0.
            ** Unallocated cells must be 0.
            */
            let mut byte_index = 0;
            let mut cell_index :usize;
            for i in 0..Self::table_size() {
                let mut next_block_id :Option<u32> = None;
                cell_index = i;

                let cell_byte_index = byte_index;
                let cell_data = u32::unpack(&block_data, &mut byte_index).expect("failed to unpack during alloc");

                //println!(" - cell {}", cell_data);

                if cell_data != 0 {
                    if cell_data & F_NOVACANCY == 0 {
                        next_block_id = Some(cell_data);
                    }
                } else {
                    // Prepare leaf block.
                    let leaf_block = self.end_block()?;
                    let mut leaf_data = vec![0u8; Z];
                    leaf_data[0] = 1;

                    // Prepare intermediate depth tables.
                    let mut next_block = leaf_block;
                    for i in 1..depth {
                        // Mark allocation on leaf table.
                        leaf_data[(i as usize) / 8] |= 1 << (i % 8);

                        // Prepare table data
                        let mut table_data = next_block.pack();
                        table_data.resize(Z, 0);

                        next_block += 1;
                        self.write_block(next_block, &table_data)?;
                    }

                    self.write_block(leaf_block, &leaf_data)?;

                    next_block_id = Some(next_block);


                    // Update cell with child table reference.
                    let pack_child = next_block.pack();
                    block_data[cell_byte_index] = pack_child[0];
                    block_data[cell_byte_index + 1] = pack_child[1];
                    block_data[cell_byte_index + 2] = pack_child[2];
                    block_data[cell_byte_index + 3] = pack_child[3];
                    write_block = true;

                    // If root table allocated last page, generate new root table at greater depth.
                    if is_root && i == Self::table_size() - 1 {
                        let parent_blocks = self.allocate(1)?;
                        let mut table_data = vec![0u8; Z];
        
                        // Update file header with new root table and depth.
                        self.file.seek(SeekFrom::Start(0))?;
                        self.file.write(&[
                            (depth as u8).pack(),
                            parent_blocks[0].pack(),
                        ].concat())?;
        
                        // Add current table to first element of new table.
                        let packed_id = block_id.pack();
                        table_data.fill(0);
                        table_data[0] = packed_id[0];
                        table_data[1] = packed_id[1];
                        table_data[2] = packed_id[2];
                        table_data[3] = packed_id[3];
                        self.write_block(parent_blocks[0], &table_data)?;
        
                        // Update current table before restarting recursion.
                        self.write_block(block_id, &block_data)?;

                        // Restart recursion with new root.
                        self.allocate_traverse(parent_blocks[0], depth + 1, 0, true, count, blocks)?;
                        return Ok(Operation::None);
                    }
                }

                // Search child table for blocks and update vacancy if requested.
                if let Some(next_block_id) = next_block_id {
                    let next_basis = Self::table_cell_offset(depth - 1, cell_index as u32, basis) * Self::pool_size() as u32;

                    //println!("@next basis {} from d {} c {} b {}", next_basis, depth, cell_index, basis);

                    match self.allocate_traverse(next_block_id, depth - 1, next_basis, false, count, blocks)? {
                        Operation::SetOccupied => {
                            write_block = true;

                            block_data[cell_byte_index + 3] |= 0x80;

                            // If last cell is marked occupied, this table is also occupied.
                            if cell_index == Self::table_size() {
                                operation = Operation::SetOccupied;
                            }
                        }
                        _ => { }
                    }
                }

                //println!(" - blocks {}", blocks.len());
                if blocks.len() == count { break; }
            }
        }
        
        // Find first cell (byte) with unset bit.
        else {
            /* ASSUMPTION
            ** This procedure should not be reachable unless table has vacancy.
            */

            // Find first unset bit in block.
            let mut byte_index = 0;
            while byte_index < block_data.len() && blocks.len() < count {
                if block_data[byte_index] != 0xFF {
                    let bit = block_data[byte_index].trailing_ones();
                    let id = basis + (byte_index * 8) as u32 + bit;

                    //println!(" - cell {} value {:02x} bit {} alloc_id {}", byte_index, block[byte_index], bit, id);

                    // Catch most common corruption case.
                    if id < 4 { panic!("invalid allocation (a < 4)"); }

                    // Add block id to output.
                    blocks.push(id);
                    //println!(" - basis {} byte_index {} byte {:02x} bit {} alloc {}", basis, byte_index, block[byte_index], bit, id);

                    // Mark block as occupied.
                    block_data[byte_index] |= 1 << bit;
                    write_block = true;
                } else {
                    byte_index += 1;
                }
            }

            // If table is fully allocated, signal parent to mark table as not vacant.
            if byte_index == Z && block_data[Z - 1] == 0xFF {
                operation = Operation::SetOccupied;
                //println!("OCCUPIED!");
            }
        }

        if write_block {
            self.write_block(block_id, &block_data)?;
        }

        Ok(operation)
    }

    fn release(&mut self, blocks:&[u32]) -> Result<(),std::io::Error>
    // Mark a set of blocks as unallocated.
    //
    {
        let mut b8 = [0u8; 1];
        let mut b32 = [0u8; 4];

        let mut block_data = [0u8; Z];

        // Read allocation table root block and depth from file.
        self.file.seek(SeekFrom::Start(0))?;
        self.file.read_exact(&mut b8)?;
        self.file.read_exact(&mut b32)?;

        let root_page = u32::unpack(&b32, &mut 0).unwrap_or_default();
        let root_depth = b8[0];

        for block in blocks {
            let block = *block;
            let mut page = root_page;
            let mut depth = root_depth as u32;

            let mut offset = 0;
            while depth > 0 {
                self.read_block(page, &mut block_data)?;

                let index = (block - offset) / Self::pool_size().pow(depth as u32) as u32;
                offset = Self::table_cell_offset(depth - 1, index as u32, offset) * Self::pool_size() as u32;

                page = u32::unpack(&block_data, &mut (4 * index as usize)).unwrap();

                depth -= 1;
            }

            self.read_block(page, &mut block_data)?;
            let bit_index = block - offset;
            let byte = bit_index / 8;
            let bit = bit_index % 8;

            block_data[byte as usize] &= !(1 << bit);
            self.write_block(page, &block_data)?;
        }

        Ok(())
    }

    fn acquire_object(&mut self, data_id:u32, length:usize) -> Result<usize, std::io::Error>
    // Allocate the next available object record.
    //
    {
        //println!("acquire_object()");

        let mut block_data = [0u8; Z];

        let mut b8 = [0u8; 1];
        let mut b32 = [0u8; 4];

        // Read allocation table root block and depth from file.
        self.file.seek(SeekFrom::Start(5))?;
        self.file.read_exact(&mut b8)?;
        let mut depth = b8[0] as u32;

        self.file.read_exact(&mut b32)?;
        let mut block_id = u32::unpack(&b32, &mut 0).unwrap_or_default();

        self.file.read_exact(&mut b32)?;
        let object_id = u32::unpack(&b32, &mut 0).unwrap_or_default() as usize;


        // Regenerate root header until range includes object_id.
        let mut range = Self::table_offset(depth + 1) as usize;
        let old_depth = depth;

        let mut root_data = vec![0u8; Z];
        while object_id > range {
            let allocation = self.allocate(1)?;

            let packed_id = block_id.pack();
            root_data[0] = packed_id[0];
            root_data[1] = packed_id[1];
            root_data[2] = packed_id[2];
            root_data[3] = packed_id[3];

            block_id = allocation[0];
            depth += 1;

            self.write_block(block_id, &root_data)?;

            range = Self::table_offset(depth + 1) as usize;
        }

        // Update header pointer and depth if changed.
        if old_depth != depth {
            let pack_depth = [ depth as u8 ];
            let pack_pointer = block_id.pack();

            self.file.seek(SeekFrom::Start(5))?;
            self.file.write(&pack_depth)?;
            self.file.write(&pack_pointer)?;
        }

        let mut basis = 0;

        // Search table for first vacant or unallocated child.
        while depth > 0 {
            /*
            ** Select child tables containing object_id until depth is 0.
            */

            //println!("basis: {}", basis);

            self.read_block(block_id, &mut block_data)?;
            let mut write_block = false;

            let cell_index = (object_id - basis) / Self::table_offset(depth) as usize;
            //println!(" - ci {}", cell_index);

            let cell_start = cell_index * 4;
            let cell_data = u32::unpack(&block_data, &mut cell_start.clone()).unwrap_or_default();

            // Allocate new page if pointer is zero.
            let child_id = if cell_data == 0 {
                let allocation = self.allocate(1)?;

                // Write new reference to table.
                let pack_block = allocation[0].pack();
                block_data[cell_start] = pack_block[0];
                block_data[cell_start + 1] = pack_block[1];
                block_data[cell_start + 2] = pack_block[2];
                block_data[cell_start + 3] = pack_block[3];

                write_block = true;

                // Populate child table.
                let mut table_data = vec![0u8; Z];

                // If child is leaf, populate next cell pointers.
                if depth == 1 {
                    let pointer_basis = Self::table_cell_offset(depth, cell_index as u32, basis as u32) + 1;

                    for i in 0..Self::table_size() {
                        let index = i * 8;
                        let pack_pointer = (pointer_basis + i as u32).pack();

                        table_data[index] = pack_pointer[0];
                        table_data[index + 1] = pack_pointer[1];
                        table_data[index + 2] = pack_pointer[2];
                        table_data[index + 3] = pack_pointer[3];
                    }
                }

                self.write_block(allocation[0], &table_data)?;

                allocation[0]
            } else {
                u32::unpack(&block_data, &mut cell_start.clone()).unwrap_or_default()
            };

            if write_block {
                self.write_block(block_id, &block_data)?;
            }

            // Update frame of reference to child table.
            block_id = child_id;
            basis = Self::table_cell_offset(depth, cell_index as u32, basis as u32) as usize;
            depth -= 1;
        }

        //println!("end basis: {}", basis);

        // Update block and header with object information.
        self.read_block(block_id, &mut block_data)?;
        
        let cell_index = (object_id - basis) / Self::table_offset(depth) as usize;
        let cell_start = cell_index * 8;
        let next_pointer = u32::unpack(&block_data, &mut cell_start.clone()).unwrap_or_default();

        //println!(" - next ptr: {}", next_pointer);

        // Update cell with data location and length.
        let pack_location = data_id.pack();
        block_data[cell_start] = pack_location[0];
        block_data[cell_start + 1] = pack_location[1];
        block_data[cell_start + 2] = pack_location[2];
        block_data[cell_start + 3] = pack_location[3];

        let pack_length = (length as u32).pack();
        block_data[cell_start + 4] = pack_length[0];
        block_data[cell_start + 5] = pack_length[1];
        block_data[cell_start + 6] = pack_length[2];
        block_data[cell_start + 7] = pack_length[3];

        // Update header with new pointer.
        let pack_pointer = next_pointer.pack();
        self.file.seek(SeekFrom::Start(10))?;
        self.file.write(&pack_pointer)?;


        self.write_block(block_id, &block_data)?;

        Ok(object_id)
    }

    fn get_object(&self, id:usize) -> Result<(u32, usize), std::io::Error>
    // Find initial block and data size of an object.
    //
    {
        let mut block_data = [0u8; Z];

        //println!("get_object()");

        let mut file = self.file.try_clone()?;

        let mut b8 = [0u8; 1];
        let mut b32 = [0u8; 4];

        // Read allocation table root block and depth from file.
        file.seek(SeekFrom::Start(5))?;
        file.read_exact(&mut b8)?;
        let mut depth = b8[0] as u32;

        file.read_exact(&mut b32)?;
        let mut block_id = u32::unpack(&b32, &mut 0).unwrap_or_default();

        let mut basis = 0;

        // Search table for first vacant or unallocated child.
        while depth > 0 {
            /*
            ** Select child tables containing object_id until depth is 0.
            */

            self.read_block(block_id, &mut block_data)?;

            let cell_index = (id - basis) / Self::table_offset(depth) as usize;

            let cell_start = cell_index * 4;
            let cell_data = u32::unpack(&block_data, &mut cell_start.clone()).unwrap_or_default();

            let child_id = if cell_data != 0 {
                u32::unpack(&block_data, &mut cell_start.clone()).unwrap_or_default()
            } else {
                return Err(std::io::Error::new(std::io::ErrorKind::NotFound, "object id not valid"));
            };

            // Update frame of reference to child table.
            block_id = child_id;
            basis = Self::table_cell_offset(depth, cell_index as u32, basis as u32) as usize;
            depth -= 1;
        }

        // Get object pointer and length from cell.
        self.read_block(block_id, &mut block_data)?;
        
        let cell_index = (id - basis) / Self::table_offset(depth) as usize;
        let mut cell_start = cell_index * 8;
        let pointer = u32::unpack(&block_data, &mut cell_start).unwrap_or_default();
        let length = u32::unpack(&block_data, &mut cell_start).unwrap_or_default();

        Ok((pointer, length as usize))
    }

    fn read_block(&self, block_id:u32, data:&mut [u8;Z]) -> Result<(),std::io::Error>
    // Read a block from file.
    //
    {
        let mut file = self.file.try_clone()?;

        file.seek(SeekFrom::Start((HEADER_SIZE + (Z * block_id as usize)) as u64))?;
        file.read_exact(data)?;

        Ok(())
    }

    fn read_block_pointer(&self, block_id:u32) -> Result<u32,std::io::Error>
    // Read the pointer element of a data block.
    //
    {
        let mut file = self.file.try_clone()?;

        let mut data = [0u8; 4];
        file.seek(SeekFrom::Start((HEADER_SIZE + (Z * block_id as usize) + Self::data_size()) as u64))?;
        file.read(&mut data)?;

        Ok(u32::unpack(&data, &mut 0).unwrap())
    }

    fn write_block(&mut self, block_id:u32, data:&[u8]) -> Result<(), std::io::Error>
    // Write a block to file.
    //
    {
        self.file.seek(SeekFrom::Start((HEADER_SIZE + (Z * block_id as usize)) as u64))?;
        self.file.write(&data[0..Z])?;

        Ok(())
    }

    fn end_block(&self) -> Result<u32, std::io::Error>
    // Id of the last block in the file.
    //
    {
        let mut file = self.file.try_clone()?;
        let index = file.seek(SeekFrom::End(0))? as usize;
        Ok((1 + (index - HEADER_SIZE) / Z) as u32)
    }

    fn table_cell_offset(depth:u32, cell:u32, basis:u32) -> u32
    {
        (Self::table_offset(depth) * cell) + basis
    }

    fn table_offset(depth:u32) -> u32
    {
        (Self::table_size() as u32).pow(depth)
    }

    fn block_count_from_size(size:usize) -> usize
    // Number of blocks required to hold a given number of bytes.
    //
    {
        (size / (Z - 4)) + ((size % (Z - 4)) != 0) as usize
    }

    const fn data_size() -> usize
    // Number of bytes of data per block.
    //
    {
        Z - 4
    }

    const fn table_size() -> usize
    // Number of elements per table block.
    //
    {
        Z / 8
    }

    const fn pool_size() -> usize
    // Number of blocks represented per pool block.
    //
    {
        Z * 8
    }
}
impl<const Z:usize> From<File> for BlockFile<Z> {
    fn from(value: File) -> Self
    {
        Self {
            file:value,
        }
    }
}
impl<const Z:usize> Into<File> for BlockFile<Z> {
    fn into(self) -> File
    {
        self.file
    }
}
